#ifndef LIB_SHARPDISP
#define LIB_SHARPDISP

#include <nrfx.h>
#include "nrf_drv_spi.h"
#include "bitmap.h"

// Sharp memory display controller API
//
// Here is an example that draws a 40x50 filled rectangle starting at 10,10
//
// #define WIDTH 400
// #define HEIGHT 240
//
// uint8_t disp_buffer[BITMAP_SIZE(WIDTH, HEIGHT)]
//
// void main() {
//   struct SharpDisp sd;
//   sharpdisp_init_default(&sd, disp_buffer, WIDTH, HEIGHT, 0xFF);
//   for (uint16_t x=0; x < 40; ++x) {
//     for (uint16_t y=0; y < 50; ++y) {
//       // bitmap_set_byte would be quicker but more complicated
//       bitmap_point(&sd.bitmap, 10 + x, 10 + y);
//     }
//   }
//
//   sharpdisp_refresh(&sd);
//   while (1);
// }


static inline void sharpdisp_init_default(
    struct SharpDisp* s,
    uint8_t* buff,
    uint16_t width,
    uint16_t height,
    uint8_t clear_byte) {
  sharpdisp_init_freq_hz(s, buff, width, height, clear_byte, 10000000);
}

// vscroll determines the vertical location of y=0 on the display.
// for eaxample, if the value is 10 then y=0 would start on physical
// line 10, lines off the bottom wrap back to the top.  The advantage
// of using this variable is that you can implement console scrolling
// without needing to use memmove() like operations to scroll the
// display up.
void sharpdisp_refresh_vscroll(struct SharpDisp* s, uint16_t vscroll);

static inline void sharpdisp_refresh(struct SharpDisp* s) {
  sharpdisp_refresh_vscroll(s, 0);
}

#endif

